package de.hawhamburg.reachability.registry;

import java.util.Random;

import com.google.gson.Gson;

import de.hawhamburg.livingplace.messaging.activemq.wrapper.LPPublisher;
import de.hawhamburg.livingplace.messaging.activemq.wrapper.LPSubscriber;
import de.hawhamburg.reachability.ProjectInfo;
import de.hawhamburg.reachability.ReachabilityWorker;
import de.hawhamburg.reachability.StopableObject;
import de.hawhamburg.reachability.conf.MainConfig;
import de.hawhamburg.reachability.registry.json.JsonRARegistryMessage;
import de.hawhamburg.reachability.registry.json.JsonRARegistryRequest;
import de.hawhamburg.reachability.util.ConnectionUtil;
import de.hawhamburg.reachability.util.OutputUtil;

/**
 * Sends register request and listens for new component registrations.<br><br>
 * 
 * This registry service sends and initial registration request over the ActiveMQ to invite all active agents to send
 * their registrations. This service establishes the systems structure via the ActiveMQ and manages their communication
 * afterwards. This may be required if there are for example two sensor agents which supply data for the same context
 * dimension. In this case, the first registration wins until the user switches the activated sensor agent by using the
 * configuration user interface.  
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public class Registry extends Thread implements StopableObject {

    /**
     * Determines whether the method {@link #shutDown()} has been called or not.
     */
    private boolean shutDown                    = false;
    
    /**
     * Publisher for register requests.
     */
    private LPPublisher requestRegistryTopic    = null;
    
    /**
     * Listener for register messages.
     */
    private LPSubscriber registryTopic          = null;
    
    /**
     * A central random object.
     */
    private Random rand                         = null;
    
    /**
     * The worker which called this property service.
     */
    private ReachabilityWorker systemAgent      = null;
    
    /**
     * Initializes a new {@link Registry} object.
     */
    public Registry() {
        // Global settings
        this.rand        = new Random();
    }
    
    /**
     * Initializes a new {@link Registry} object with an {@link ReachabilityWorker}.
     * 
     * @param worker    The {@link ReachabilityWorker} which started this registry service.
     */
    public Registry(final ReachabilityWorker worker) {
        // Global settings
        this.rand        = new Random();
        this.systemAgent = worker;
    }

    /**
     * @see StopableObject#isActive()
     * 
     * @return      <code>true</code> if {@link #shutDown()} wasn't called, otherwise <code>true</code>.
     */
    public final boolean isActive() {
        return !this.shutDown;
    }

    /**
     * @see StopableObject#isActive()
     */
    public final void shutDown() {
        this.shutDown = true;
    }
    
    /**
     * Starts the {@link Registry} service. Initially it sends a registry request and handles registry messages 
     * afterwards, until the {@link #shutDown()} method has been called. 
     */
    public final void run() {
        // Initialize connections
        init();
        
        // Send initial registry request to active sensors
        sendInitialRegistryRequest();
        
        // Wait for registrations and departures
        while (this.isActive()) {
            handleRegistryMessage();
        }
        
        // End system and free results
        end();
    }
    
    /**
     * Open all required ActiveMQ connections.
     */
    private void init() {
        // Fetch topic names
        String registryRequestTopicName = MainConfig.get("REACHABILITY_REGISTRY_REQUEST_TOPIC");
        String registryTopicName        = MainConfig.get("REACHABILITY_REGISTRY_TOPIC");
        
        // Set up subscribers/publishers
        this.requestRegistryTopic       = ConnectionUtil.getPublisher(registryRequestTopicName, this);
        this.registryTopic              = ConnectionUtil.getSubscriber(registryTopicName, this);
    }
    
    /**
     * Sends an initial {@link JsonRARegistryRequest} message to invite all active agents.
     */
    private void sendInitialRegistryRequest() {
        JsonRARegistryRequest req = new JsonRARegistryRequest();
        
        req.Id      = getID();
        req.Version = getVersion();
        req.Remark  = "Send your registrations now!";
        
        OutputUtil.info("  Sending initial register request and waiting for answers");
        ConnectionUtil.setMessageOnTopic(requestRegistryTopic, req);
    }
    
    /**
     * Handle incoming {@link JsonRARegistryMessage} messages which indicates that an agent would like to register or
     * unregister itself to/from the system.
     */
    private void handleRegistryMessage() {
        String json = ConnectionUtil.getMessageOnTopic(registryTopic);
        if (json != null) {
            Gson g = new Gson();
            JsonRARegistryMessage msgObj = g.fromJson(json, JsonRARegistryMessage.class);
            
            // Delegate to reachability agent
            switch(msgObj.MessageType) {
                case UNREGISTER:
                    this.systemAgent.unregister(msgObj.AgentType, msgObj.Name, msgObj.Dimension);
                    break;
                case REGISTER:
                default:
                    this.systemAgent.register(msgObj.AgentType, msgObj.Name, msgObj.Dimension);
                    break;
                    
            }
        }
    }
    
    /**
     * Gracefully ends the system by disconnecting all open ActiveMQ connections.
     */
    private void end() {
        if (requestRegistryTopic != null) {
            requestRegistryTopic.disconnect();
        }
        if (registryTopic != null) {
            registryTopic.disconnect();
        }
    }
    
    /**
     * Returns an service ID.
     * 
     * @return              The service ID containing the {@link ProjectInfo#REACHABILITY_AGENT_WORKER_NAME}, the 
     *                      services name and a pseudo random number.
     */
    private String getID() {
        return ProjectInfo.REACHABILITY_AGENT_WORKER_NAME + "_Registry_" + this.rand.nextInt(Integer.MAX_VALUE);
    }
    
    /**
     * Returns the agents version.
     * 
     * @return              The agents version as {@link ProjectInfo#REACHABILITY_AGENT_VERSION}.
     */
    private String getVersion() {
        return ProjectInfo.REACHABILITY_AGENT_VERSION;
    }
}
