package de.hawhamburg.reachability.util;

import java.net.UnknownHostException;

import javax.jms.JMSException;

import com.google.gson.Gson;
import com.google.gson.JsonIOException;
import com.mongodb.MongoException;

import de.hawhamburg.livingplace.messaging.activemq.wrapper.ConnectionSettings;
import de.hawhamburg.livingplace.messaging.activemq.wrapper.LPPublisher;
import de.hawhamburg.livingplace.messaging.activemq.wrapper.LPSubscriber;
import de.hawhamburg.reachability.StopableObject;

/**
 * Utility functions related to Active MQ connections.<br><br>
 * 
 * This utility class provides some fancy utility methods related to the Active MQ message interaction.
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public final class ConnectionUtil {

    /**
     * Waiting time before reconnect in milliseconds.
     */
    private static final long WAIT_BEFORE_RECONNECT = 10000;
    
    /**
     * The timeout after which the ActiveMQ topic will be freed on read.
     */
    private static final long AMQ_READ_TIMEOUT      = 1000;
    
    /**
     * Holds the connection settings. Will be loaded only once in a programs lifetime. For changes, the program needs 
     * to be restarted.
     */
    private static ConnectionSettings conSettings   = null;

    /**
     * Prevent instances of this utility class.
     */
    private ConnectionUtil() { }
    
    /**
     * Returns an {@link LPSubscriber} object related to a certain topic. The second parameter should be the calling 
     * object, if it is kind of {@link StopableObject}. To ensure, that the connection attempt will be stopped, of the 
     * calling object was shut down.
     * 
     * @param topic     The topics name as {@link String}.
     * @param obj       The calling object, if it's a {@link StopableObject} (<code>null</code> instead).
     * @return          An {@link LPSubscriber} object to read messages from this topic.
     */
    public static LPSubscriber getSubscriber(final String topic,
                                             final StopableObject obj) {
        if (topic == null) {
            return null;
        }

        if (conSettings == null) {
            // Define connection settings
            conSettings             = new ConnectionSettings();
            conSettings.amq_ip      = de.hawhamburg.lifecycle.data.Config.ACTIVEMQIP;
            conSettings.mongo_ip    = de.hawhamburg.lifecycle.data.Config.MONGODBIP;
        }
        
        LPSubscriber subscr = null;
        boolean connected = false;
        do {
            try {
                connected   = false;
                subscr      = new LPSubscriber(topic, conSettings);
                connected   = true;
            } catch (JMSException e) {
                System.err.println("connection fail, waiting " + WAIT_BEFORE_RECONNECT + "ms");
                try {
                    Thread.sleep(WAIT_BEFORE_RECONNECT);
                } catch (InterruptedException e1) {
                    OutputUtil.error("[InterruptedException]" + e.getMessage());
                }
            }
        } while (!connected && (obj == null || obj.isActive()));
        
        return subscr;
    }
    
    /**
     * Returns an {@link LPPublisher} object related to a certain topic. The second parameter should be the calling 
     * object, if it is kind of {@link StopableObject}. To ensure, that the connection attempt will be stopped, of the 
     * calling object was shut down.
     * 
     * @param topic     The topics name as {@link String}.
     * @param obj       The calling object, if it's a {@link StopableObject} (<code>null</code> instead).
     * @return          An {@link LPPublisher} object to set messages on this topic.
     */
    public static LPPublisher getPublisher(final String topic,
                                           final StopableObject obj) {
        if (topic == null) {
            return null;
        }
        
        if (conSettings == null) {
            // Define connection settings
            conSettings             = new ConnectionSettings();
            conSettings.amq_ip      = de.hawhamburg.lifecycle.data.Config.ACTIVEMQIP;
            conSettings.mongo_ip    = de.hawhamburg.lifecycle.data.Config.MONGODBIP;
        }
        
        LPPublisher pub = null;
        boolean connected = false;
        do {
            try {
                connected   = false;
                pub         = new LPPublisher(topic, conSettings);
                connected   = true;
            } catch (UnknownHostException e) {
                System.err.println("host fail, waiting " + WAIT_BEFORE_RECONNECT + "ms");
                try {
                    Thread.sleep(WAIT_BEFORE_RECONNECT);
                } catch (InterruptedException e1) {
                    OutputUtil.error("[InterruptedException]" + e.getMessage());
                }
            } catch (MongoException e) {
                System.err.println("mongoDb fail, waiting " + WAIT_BEFORE_RECONNECT + "ms");
                try {
                    Thread.sleep(WAIT_BEFORE_RECONNECT);
                } catch (InterruptedException e1) {
                    OutputUtil.error("[InterruptedException]" + e.getMessage());
                }
            } catch (JMSException e) {
                System.err.println("connection fail, waiting " + WAIT_BEFORE_RECONNECT + "ms");
                try {
                    Thread.sleep(WAIT_BEFORE_RECONNECT);
                } catch (InterruptedException e1) {
                    OutputUtil.error("[InterruptedException]" + e.getMessage());
                }
            }
        } while (!connected && (obj == null || obj.isActive()));
        
        return pub;
    }

    /**
     * Checks if there is a message on the topic of the {@link LPSubscriber} object and returns it as {@link String} on 
     * success (<code>null</code> instead). 
     * 
     * @param topic     An {@link LPSubscriber} object, which shall be checked for new messages.
     * @return          A message as {@link String} on success, otherwise <code>null</code>.
     */
    public static String getMessageOnTopic(final LPSubscriber topic) {
        if (topic == null) {
            return null;
        }
        
        String message = null;
        try {
            message = topic.subscribeBlockingWithTimeout(AMQ_READ_TIMEOUT);
            
            if (message != null) {
                return message;
            }
        } catch (InterruptedException e) {
            OutputUtil.error("[InterruptedException] " + e.getMessage());
        } catch (JMSException e) {
            OutputUtil.error("[JMSException] " + e.getMessage());
        } catch (JsonIOException e) {
            OutputUtil.error("[JsonIOException] Gson failed to read: " + message);
        }
        
        return null;
    }
    
    /**
     * Tries to send a JSON message object on a certain topic using an {@link LPPublisher} object.
     * 
     * @param topic     An {@link LPPublisher} object to send on.
     * @param jsonMsg   A JSON message object.
     */
    public static void setMessageOnTopic(final LPPublisher topic,
                                         final Object jsonMsg) {
        if (topic != null) {
            Gson g = new Gson();
            topic.publish(g.toJson(jsonMsg));
        }
    }
}
