package de.hawhamburg.reachability.util;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;

/**
 * Utility functions related to external file handling.<br><br>
 * 
 * This utility class provides some fancy methods related to external file modifications.
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public final class ExternFileUtil {

    /**
     * Prevent instances of this utility class.
     */
    private ExternFileUtil() { }
    
    /**
     * Checks whether the supplied file path exists or not. 
     * 
     * @param path      The path referred to the file which existence shall be checked.
     * @return          <code>true</code> if the supplied path refers to an existing file, otherwise <code>false</code>.
     */
    public static boolean exists(final String path) {
        if (path == null) {
            return false;
        }
        return new File(path).exists();
    }
    
    /**
     * Check whether the file is accessible for read operations or not. This includes an existence and an is file check.
     * 
     * @param path      The path referred to the file of which the accessibility shall be checked. 
     * @return          <code>true</code> if the supplied path refers to an accessible file, otherwise 
     *                  <code>false</code>.
     */
    public static boolean isAccessible(final String path) {
        return exists(path) && isFile(path) && new File(path).canRead();
    }
    
    /**
     * Checks whether the supplied path refers to a file or not. 
     * 
     * @param path      The path which shall be checked.
     * @return          <code>true</code> if the supplied path refers to a file, otherwise <code>false</code>.
     */
    public static boolean isFile(final String path) {
        return isFile(path, "*");
    }
    
    /**
     * Checks whether the supplied path refers to a file of a certain type or not.
     * 
     * @param path      The path which shall be checked.
     * @param suffix    The suffix of this file.
     * @return          <code>true</code> if the supplied path refers to a file, otherwise <code>false</code>.
     */
    public static boolean isFile(final String path,
                                 final String suffix) {
        boolean result = exists(path) && new File(path).isFile();
        
        if (result && suffix != "*") {
            result = path.endsWith("." + suffix);
        }
        
        return result;
    }
    
    /**
     * Returns the content of the supplied file as list of lines. If the file doesn't exist, <code>null</code> will be 
     * returned instead.
     * 
     * @param path      The file which content shall be returned as list of lines.
     * @return          A {@link List} of lines as {@link String} representing the supplied files content.
     */
    public static List<String> getContentLines(final String path) {
        List<String> result = new ArrayList<String>();
        if (exists(path)) {
            try {
                FileInputStream fstream = new FileInputStream(path);
                DataInputStream in      = new DataInputStream(fstream);
                BufferedReader br       = new BufferedReader(new InputStreamReader(in));
                String strLine;
                while ((strLine = br.readLine()) != null) {
                    result.add(strLine);
                }
                in.close();
            } catch (UnknownHostException e) {
                OutputUtil.error("[UnknownHostException] "  + e.getMessage());
            } catch (IOException e) {
                OutputUtil.error("[IOException] "           + e.getMessage());
            }
            
            return result;
        }
        return null;
    }
    
    /**
     * Removes the file referred to by the supplied file path if it's existing.
     * 
     * @param path      The path pointing to the file which shall be removed.
     * @return          <code>true</code> on success, otherwise <code>false</code>.
     */
    public static boolean remove(final String path) {
        boolean result = true;
        if (exists(path)) {
            result = false;
            File f = new File(path);
            if (f.canWrite()) {
                if (!f.isDirectory()) {
                    result = f.delete();
                }
            }
        }
        return result;
    }
    
    /**
     * Creates a new empty file with the supplied path name, if the path doesn't refer to an existing file.  
     * 
     * @param path      The path to the new file.
     * @return          <code>true</code> on success, otherwise <code>false</code>.
     */
    public static boolean createEmpty(final String path) {
        boolean result = false;
        if (!exists(path)) {
            File f = new File(path);
            try {
                result = f.createNewFile();
            } catch (UnknownHostException e) {
                result = false;
            } catch (IOException e) {
                result = false;
            }
        }
        return result;
    }
    
    /**
     * Writes the given content to the file referred to by the supplied path name.
     * 
     * @param path      The path to the file to which the supplied content shall be written.
     * @param content   The content which shall be written to the given file.
     * @return          <code>true</code> on success, otherwise <code>false</code>.
     */
    public static boolean write(final String path,
                                final String content) {
        boolean result = false;
        if (exists(path)) {
            try {
                FileWriter fstream = new FileWriter(path);
                BufferedWriter out = new BufferedWriter(fstream);
                out.write(content);
                out.close();
                result = true;
            } catch (UnknownHostException e) {
                result = false;
            } catch (IOException e) {
                result = false;
            }
        }
        return result;
    }
    
    /**
     * Creates a new file which is referred to be the given path and writes the supplied content into it, by using 
     * {@link #createEmpty(String)}.
     * 
     * @param path      The path the new file which shall be created.
     * @param content   The content which shall be written to the new file.
     * @return          <code>true</code>, on success, otherwise <code>false</code>.
     */
    public static boolean createAndWrite(final String path,
                                         final String content) {
        boolean result = false;
        if (createEmpty(path)) {
            result = write(path, content);
        }
        return result;
    }
}
